@@ -3,121 +3,123 @@
 {-# LANGUAGE MagicHash         #-}
 {-# LANGUAGE OverloadedStrings #-}
 
+------------------------------------------------------------------------------
 module Snap.Internal.Parsing where
-
 ------------------------------------------------------------------------------
-import           Blaze.ByteString.Builder
-import           Control.Applicative
-import           Control.Arrow                 (first, second)
-import           Control.Monad
-import           Data.Attoparsec.Char8
-import           Data.Bits
-import           Data.ByteString.Char8         (ByteString)
-import qualified Data.ByteString.Char8         as S
-import           Data.ByteString.Internal      (c2w, w2c)
-import qualified Data.ByteString.Lazy.Char8    as L
-import           Data.CaseInsensitive          (CI)
-import qualified Data.CaseInsensitive          as CI
-import           Data.Char                     hiding (isDigit, isSpace)
-import           Data.Int
-import           Data.List                     (intersperse)
-import           Data.Map                      (Map)
-import qualified Data.Map                      as Map
-import           Data.Maybe
-import           Data.Monoid
-import           Data.Word
-import           GHC.Exts
-import           GHC.Word                      (Word8 (..))
-import           Prelude                       hiding (head, take, takeWhile)
+import           Control.Applicative              (Alternative ((<|>)), Applicative ((*>), (<*), pure), liftA2, (<$>))
+import           Control.Arrow                    (first, second)
+import           Control.Monad                    (Monad (return), MonadPlus (mzero), liftM, when)
+import           Data.Attoparsec.ByteString.Char8 (IResult (Done, Fail, Partial), Parser, Result, anyChar, char, choice, decimal, endOfInput, feed, inClass, isDigit, isSpace, letter_ascii, many', match, option, parse, satisfy, skipSpace, skipWhile, string, take, takeTill, takeWhile)
+import qualified Data.Attoparsec.ByteString.Char8 as AP
+import           Data.Bits                        (Bits ((.&.), (.|.), unsafeShiftL))
+import           Data.ByteString.Builder          (Builder, byteString, char8, toLazyByteString, word8)
+import           Data.ByteString.Char8            (ByteString)
+import qualified Data.ByteString.Char8            as S
+import           Data.ByteString.Internal         (c2w, w2c)
+import qualified Data.ByteString.Lazy.Char8       as L
+import           Data.CaseInsensitive             (CI)
+import qualified Data.CaseInsensitive             as CI (mk)
+import           Data.Char                        (Char, intToDigit, isAlpha, isAlphaNum, isAscii, isControl, isHexDigit, ord)
+import           Data.Int                         (Int64)
+import           Data.List                        (concat, intercalate, intersperse)
+import           Data.Map                         (Map)
+import qualified Data.Map                         as Map (empty, insertWith', toList)
+import           Data.Maybe                       (Maybe (..), maybe)
+import           Data.Monoid                      (Monoid (mconcat, mempty), (<>))
+import           Data.Word                        (Word8)
+import           GHC.Exts                         (Int (I#), uncheckedShiftRL#, word2Int#)
+import           GHC.Word                         (Word8 (..))
+import           Prelude                          (Bool (..), Either (..), Enum (fromEnum, toEnum), Eq (..), Num (..), Ord (..), String, and, any, concatMap, elem, error, filter, flip, foldr, fst, id, map, not, otherwise, show, snd, ($), ($!), (&&), (++), (.), (||))
+import           Snap.Internal.Http.Types         (Cookie (Cookie))
 ------------------------------------------------------------------------------
-import           Snap.Internal.Http.Types
-import           Snap.Internal.Parsing.FastSet (FastSet)
-import qualified Snap.Internal.Parsing.FastSet as FS
 
 
 ------------------------------------------------------------------------------
 {-# INLINE fullyParse #-}
 fullyParse :: ByteString -> Parser a -> Either String a
-fullyParse s p =
+fullyParse = fullyParse' parse feed
+
+{-# INLINE (<?>) #-}
+(<?>) :: Parser a -> String -> Parser a
+(<?>) a !b = (AP.<?>) a b
+infix 0 <?>
+
+------------------------------------------------------------------------------
+{-# INLINE fullyParse' #-}
+fullyParse' :: (Parser a -> ByteString -> Result a)
+            -> (Result a -> ByteString -> Result a)
+            -> ByteString
+            -> Parser a
+            -> Either String a
+fullyParse' parseFunc feedFunc s p =
     case r' of
-      (Fail _ _ e) -> Left e
-      (Partial _)  -> Left "parse failed"
+      (Fail _ context e) -> Left $ concat [ "Parsing "
+                                          , intercalate "/" context
+                                          , ": "
+                                          , e
+                                          , "."
+                                          ]
+      (Partial _)  -> Left "parse failed"  -- expected to be impossible
       (Done _ x)   -> Right x
   where
-    r  = parse p s
-    r' = feed r ""
+    r  = parseFunc p s
+    r' = feedFunc r ""
 
+------------------------------------------------------------------------------
+-- Parsers for different tokens in an HTTP request.
 
 ------------------------------------------------------------------------------
 parseNum :: Parser Int64
 parseNum = decimal
 
 
-------------------------------------------------------------------------------
--- | Parsers for different tokens in an HTTP request.
-sp, digit, letter :: Parser Char
-sp       = char ' '
-digit    = satisfy isDigit
-letter   = satisfy isAlpha
-
-
 ------------------------------------------------------------------------------
 untilEOL :: Parser ByteString
-untilEOL = takeWhile notend
+untilEOL = takeWhile notend <?> "untilEOL"
   where
     notend c = not $ c == '\r' || c == '\n'
 
 
 ------------------------------------------------------------------------------
 crlf :: Parser ByteString
-crlf = string "\r\n"
+crlf = string "\r\n" <?> "crlf"
 
 
 ------------------------------------------------------------------------------
-generateFS :: (Word8 -> Bool) -> FastSet
-generateFS f = FS.fromList $ filter f [0..255]
+toTable :: (Char -> Bool) -> (Char -> Bool)
+toTable f = inClass $ filter f $ map w2c [0..255]
+{-# INLINE toTable #-}
 
 
 ------------------------------------------------------------------------------
--- | Parser for zero or more spaces.
-spaces :: Parser [Char]
-spaces = many sp
+skipFieldChars :: Parser ()
+skipFieldChars = skipWhile isFieldChar
 
 
 ------------------------------------------------------------------------------
-pSpaces :: Parser ByteString
-pSpaces = takeWhile isSpace
-
-
-------------------------------------------------------------------------------
-fieldChars :: Parser ByteString
-fieldChars = takeWhile isFieldChar
+isFieldChar :: Char -> Bool
+isFieldChar = toTable f
   where
-    isFieldChar = flip FS.memberChar fieldCharSet
-
-
-------------------------------------------------------------------------------
-fieldCharSet :: FastSet
-fieldCharSet = generateFS f
-  where
-    f d = let c = (toEnum $ fromEnum d)
-          in (isDigit c) || (isAlpha c) || c == '-' || c == '_'
+    f c = (isDigit c) || (isAlpha c) || c == '-' || c == '_'
 
 
 ------------------------------------------------------------------------------
 -- | Parser for request headers.
 pHeaders :: Parser [(ByteString, ByteString)]
-pHeaders = many header
+pHeaders = many' header <?> "headers"
   where
+    --------------------------------------------------------------------------
+    slurp p = fst <$> match p
+
     --------------------------------------------------------------------------
     header            = {-# SCC "pHeaders/header" #-}
                         liftA2 (,)
                             fieldName
-                            (char ':' *> spaces *> contents)
+                            (char ':' *> skipSpace *> contents)
 
     --------------------------------------------------------------------------
     fieldName         = {-# SCC "pHeaders/fieldName" #-}
-                        liftA2 S.cons letter fieldChars
+                        slurp (letter_ascii *> skipFieldChars)
 
     --------------------------------------------------------------------------
     contents          = {-# SCC "pHeaders/contents" #-}
@@ -131,7 +133,7 @@ pHeaders = many header
 
     --------------------------------------------------------------------------
     leadingWhiteSpace = {-# SCC "pHeaders/leadingWhiteSpace" #-}
-                        takeWhile1 isLeadingWS
+                        skipWhile1 isLeadingWS
 
     --------------------------------------------------------------------------
     continuation      = {-# SCC "pHeaders/continuation" #-}
@@ -139,6 +141,9 @@ pHeaders = many header
                                (leadingWhiteSpace *> pure ' ')
                                contents
 
+    --------------------------------------------------------------------------
+    skipWhile1 f = satisfy f *> skipWhile f
+
 
 ------------------------------------------------------------------------------
 -- unhelpfully, the spec mentions "old-style" cookies that don't have quotes
@@ -151,17 +156,17 @@ pWord = pQuotedString <|> (takeWhile (/= ';'))
 pQuotedString :: Parser ByteString
 pQuotedString = q *> quotedText <* q
   where
-    quotedText = (S.concat . reverse) <$> f []
+    quotedText = (S.concat . L.toChunks . toLazyByteString) <$> f mempty
 
     f soFar = do
         t <- takeWhile qdtext
-        let soFar' = t:soFar
+        let soFar' = soFar <> byteString t
         -- RFC says that backslash only escapes for <">
-        choice [ string "\\\"" *> f ("\"" : soFar')
+        choice [ string "\\\"" *> f (soFar' <> char8 '"')
                , pure soFar' ]
 
-    q      = char '\"'
-    qdtext = matchAll [ isRFCText, (/= '\"'), (/= '\\') ]
+    q      = char '"'
+    qdtext = matchAll [ isRFCText, (/= '"'), (/= '\\') ]
 
 
 ------------------------------------------------------------------------------
@@ -180,7 +185,7 @@ matchAll x c = and $ map ($ c) x
 pAvPairs :: Parser [(ByteString, ByteString)]
 pAvPairs = do
     a <- pAvPair
-    b <- many (pSpaces *> char ';' *> pSpaces *> pAvPair)
+    b <- many' (skipSpace *> char ';' *> skipSpace *> pAvPair)
     return $! a:b
 
 
@@ -188,17 +193,19 @@ pAvPairs = do
 {-# INLINE pAvPair #-}
 pAvPair :: Parser (ByteString, ByteString)
 pAvPair = do
-    key <- pToken <* pSpaces
-    val <- liftM trim (option "" $ char '=' *> pSpaces *> pWord)
+    key <- pToken <* skipSpace
+    val <- liftM trim (option "" $ char '=' *> skipSpace *> pWord)
     return $! (key, val)
 
 
 ------------------------------------------------------------------------------
 pParameter :: Parser (ByteString, ByteString)
-pParameter = do
-    key <- pToken <* pSpaces
-    val <- liftM trim (char '=' *> pSpaces *> pWord)
-    return $! (trim key, val)
+pParameter = parser <?> "pParameter"
+  where
+    parser = do
+        key <- pToken <* skipSpace
+        val <- liftM trim (char '=' *> skipSpace *> pWord)
+        return $! (trim key, val)
 
 
 ------------------------------------------------------------------------------
@@ -209,24 +216,27 @@ trim = snd . S.span isSpace . fst . S.spanEnd isSpace
 
 ------------------------------------------------------------------------------
 pValueWithParameters :: Parser (ByteString, [(CI ByteString, ByteString)])
-pValueWithParameters = do
-    value  <- liftM trim (pSpaces *> takeWhile (/= ';'))
-    params <- many pParam
-    return (value, map (first CI.mk) params)
-
+pValueWithParameters = parser <?> "pValueWithParameters"
   where
-    pParam = pSpaces *> char ';' *> pSpaces *> pParameter
+    parser = do
+        value  <- liftM trim (skipSpace *> takeWhile (/= ';'))
+        params <- many' pParam
+        endOfInput
+        return (value, map (first CI.mk) params)
+    pParam = skipSpace *> char ';' *> skipSpace *> pParameter
 
 
 ------------------------------------------------------------------------------
 pContentTypeWithParameters :: Parser ( ByteString
                                      , [(CI ByteString, ByteString)] )
-pContentTypeWithParameters = do
-    value  <- liftM trim (pSpaces *> takeWhile (not . isSep))
-    params <- many (pSpaces *> satisfy isSep *> pSpaces *> pParameter)
-    return $! (value, map (first CI.mk) params)
-
+pContentTypeWithParameters = parser <?> "pContentTypeWithParameters"
   where
+    parser = do
+        value  <- liftM trim (skipSpace *> takeWhile (not . isSep))
+        params <- many' (skipSpace *> satisfy isSep *> skipSpace *> pParameter)
+        endOfInput
+        return $! (value, map (first CI.mk) params)
+
     isSep c = c == ';' || c == ','
 
 
@@ -239,12 +249,7 @@ pToken = takeWhile isToken
 ------------------------------------------------------------------------------
 {-# INLINE isToken #-}
 isToken :: Char -> Bool
-isToken c = FS.memberChar c tokenTable
-
-
-------------------------------------------------------------------------------
-tokenTable :: FastSet
-tokenTable = generateFS (f . toEnum . fromEnum)
+isToken = toTable f
   where
     f = matchAll [ isAscii
                  , not . isControl
@@ -293,7 +298,7 @@ pUrlEscaped = do
     percentEncoded !l = do
         hx <- take 2
         when (S.length hx /= 2 || (not $ S.all isHexDigit hx)) $
-             fail "bad hex in url"
+             mzero
 
         let code = w2c ((unsafeFromHex hx) :: Word8)
         nextChunk $ l . ((S.singleton code) :)
@@ -318,49 +323,74 @@ pUrlEscaped = do
 -- reserved purposes may be used unencoded within a URL."
 
 
+
+
 ------------------------------------------------------------------------------
--- | Decodes an URL-escaped string (see
+-- | Decode an URL-escaped string (see
 -- <http://tools.ietf.org/html/rfc2396.html#section-2.4>)
+--
+-- Example:
+--
+-- @
+-- ghci> 'urlDecode' "1+attoparsec+%7e%3d+3+*+10%5e-2+meters"
+-- Just "1 attoparsec ~= 3 * 10^-2 meters"
+-- @
 urlDecode :: ByteString -> Maybe ByteString
 urlDecode = parseToCompletion pUrlEscaped
 {-# INLINE urlDecode #-}
 
 
 ------------------------------------------------------------------------------
--- | URL-escapes a string (see
+-- | URL-escape a string (see
 -- <http://tools.ietf.org/html/rfc2396.html#section-2.4>)
+--
+-- Example:
+--
+-- @
+-- ghci> 'urlEncode' "1 attoparsec ~= 3 * 10^-2 meters"
+-- "1+attoparsec+%7e%3d+3+*+10%5e-2+meters"
+-- @
 urlEncode :: ByteString -> ByteString
-urlEncode = toByteString . urlEncodeBuilder
+urlEncode = S.concat . L.toChunks . toLazyByteString . urlEncodeBuilder
 {-# INLINE urlEncode #-}
 
 
 ------------------------------------------------------------------------------
--- | URL-escapes a string (see
+-- | URL-escape a string (see
 -- <http://tools.ietf.org/html/rfc2396.html#section-2.4>) into a 'Builder'.
+--
+-- Example:
+--
+-- @
+-- ghci> import "Data.ByteString.Builder"
+-- ghci> 'toLazyByteString' . 'urlEncodeBuilder' $ "1 attoparsec ~= 3 * 10^-2 meters"
+-- "1+attoparsec+%7e%3d+3+*+10%5e-2+meters"
+-- @
 urlEncodeBuilder :: ByteString -> Builder
 urlEncodeBuilder = go mempty
   where
     go !b !s = maybe b' esc (S.uncons y)
       where
-        (x,y)     = S.span (flip FS.memberChar urlEncodeTable) s
-        b'        = b `mappend` fromByteString x
+        (x,y)     = S.span urlEncodeClean s
+        b'        = b <> byteString x
         esc (c,r) = let b'' = if c == ' '
-                                then b' `mappend` fromWord8 (c2w '+')
-                                else b' `mappend` hexd c
+                                then b' <> char8 '+'
+                                else b' <> hexd c
                     in go b'' r
 
 
 ------------------------------------------------------------------------------
-urlEncodeTable :: FastSet
-urlEncodeTable = generateFS f
+urlEncodeClean :: Char -> Bool
+urlEncodeClean = toTable f
   where
-    f c = any ($ (w2c c)) [\x -> isAscii x && isAlphaNum x,
-                          flip elem [ '$', '_', '-', '.', '!'
-                                    , '*' , '\'', '(', ')', ',' ] ]
+    f c = any ($ c) [\c' -> isAscii c' && isAlphaNum c'
+                    , flip elem [ '$', '_', '-', '.', '!'
+                                , '*' , '\'', '(', ')', ',' ]]
+
 
 ------------------------------------------------------------------------------
 hexd :: Char -> Builder
-hexd c0 = fromWord8 (c2w '%') `mappend` fromWord8 hi `mappend` fromWord8 low
+hexd c0 = char8 '%' <> word8 hi <> word8 low
   where
     !c        = c2w c0
     toDigit   = c2w . intToDigit
@@ -381,7 +411,14 @@ finish x           = x
                     ---------------------------------------
 
 ------------------------------------------------------------------------------
--- | Parses a string encoded in @application/x-www-form-urlencoded@ format.
+-- | Parse a string encoded in @application/x-www-form-urlencoded@ < http://en.wikipedia.org/wiki/POST_%28HTTP%29#Use_for_submitting_web_forms format>.
+--
+-- Example:
+--
+-- @
+-- ghci> 'parseUrlEncoded' "Name=John+Doe&Name=Jane+Doe&Age=23&Formula=a+%2B+b+%3D%3D+13%25%21"
+-- 'Data.Map.fromList' [("Age",["23"]),("Formula",["a + b == 13%!"]),("Name",["John Doe","Jane Doe"])]
+-- @
 parseUrlEncoded :: ByteString -> Map ByteString [ByteString]
 parseUrlEncoded s = foldr ins Map.empty decoded
 
@@ -416,22 +453,46 @@ parseUrlEncoded s = foldr ins Map.empty decoded
 
 
 ------------------------------------------------------------------------------
+-- | Like 'printUrlEncoded', but produces a 'Builder' instead of a
+-- 'ByteString'. Useful for constructing a large string efficiently in
+-- a single step.
+--
+-- Example:
+--
+-- @
+-- ghci> import "Data.Map"
+-- ghci> import "Data.Monoid"
+-- ghci> import "Data.ByteString.Builder"
+-- ghci> let bldr = 'buildUrlEncoded' ('Data.Map.fromList' [("Name", ["John Doe"]), ("Age", ["23"])])
+-- ghci> 'toLazyByteString' $ 'byteString' "http://example.com/script?" <> bldr
+-- "http://example.com/script?Age=23&Name=John+Doe"
+-- @
 buildUrlEncoded :: Map ByteString [ByteString] -> Builder
 buildUrlEncoded m = mconcat builders
   where
-    builders        = intersperse (fromWord8 $ c2w '&') $
+    builders        = intersperse (char8 '&') $
                       concatMap encodeVS $ Map.toList m
 
     encodeVS (k,vs) = map (encodeOne k) vs
 
     encodeOne k v   = mconcat [ urlEncodeBuilder k
-                              , fromWord8 $ c2w '='
+                              , char8 '='
                               , urlEncodeBuilder v ]
 
 
 ------------------------------------------------------------------------------
+-- | Given a collection of key-value pairs with possibly duplicate
+-- keys (represented as a 'Data.Map.Map'), construct a string in
+-- @application/x-www-form-urlencoded@ format.
+--
+-- Example:
+--
+-- @
+-- ghci> 'printUrlEncoded' ('Data.Map.fromList' [("Name", ["John Doe"]), ("Age", ["23"])])
+-- "Age=23&Name=John+Doe"
+-- @
 printUrlEncoded :: Map ByteString [ByteString] -> ByteString
-printUrlEncoded = toByteString . buildUrlEncoded
+printUrlEncoded = S.concat . L.toChunks . toLazyByteString . buildUrlEncoded
 
 
                              --------------------
@@ -461,10 +522,6 @@ parseCookie = parseToCompletion pCookies
                             -- utility functions --
                             -----------------------
 
-------------------------------------------------------------------------------
-strictize :: L.ByteString -> ByteString
-strictize = S.concat . L.toChunks
-
 ------------------------------------------------------------------------------
 unsafeFromHex :: (Enum a, Num a, Bits a) => ByteString -> a
 unsafeFromHex = S.foldl' f 0
@@ -485,8 +542,16 @@ unsafeFromHex = S.foldl' f 0
 
 
 ------------------------------------------------------------------------------
-unsafeFromInt :: (Enum a, Num a, Bits a) => ByteString -> a
-unsafeFromInt = S.foldl' f 0
+-- Note: only works for nonnegative naturals
+unsafeFromNat :: (Enum a, Num a, Bits a) => ByteString -> a
+unsafeFromNat = S.foldl' f 0
   where
+    zero = ord '0'
     f !cnt !i = cnt * 10 + toEnum (digitToInt i)
-{-# INLINE unsafeFromInt #-}
+
+    digitToInt c = if d >= 0 && d <= 9
+                     then d
+                     else error $ "bad digit: '" ++ [c] ++ "'"
+      where
+        !d = ord c - zero
+{-# INLINE unsafeFromNat #-}
