@@ -1,45 +1,48 @@
 {-# LANGUAGE BangPatterns     #-}
+{-# LANGUAGE CPP              #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE Rank2Types       #-}
-module Snap.Internal.Routing where
-
-
+module Snap.Internal.Routing
+  ( Route(..)
+  , pRoute
+  , route
+  , routeEarliestNC
+  , routeHeight
+  , routeLocal
+  , splitPath
+  ) where
 ------------------------------------------------------------------------------
-import           Control.Applicative ((<|>))
-import           Data.ByteString (ByteString)
+import           Control.Applicative      ((<|>))
+import           Data.ByteString          (ByteString)
+import qualified Data.ByteString          as B (head, intercalate, length, null, pack, splitWith, tail)
 import           Data.ByteString.Internal (c2w)
-import qualified Data.ByteString as B
-import           Data.Monoid
-import           Data.HashMap.Strict (HashMap)
-import qualified Data.HashMap.Strict as H
-import qualified Data.Map as Map
+import           Data.HashMap.Strict      (HashMap)
+import qualified Data.HashMap.Strict      as H (elems, empty, fromList, lookup, unionWith)
+import qualified Data.Map                 as Map (empty, insertWith, unionWith)
+import           Snap.Internal.Core       (MonadSnap, getRequest, getsRequest, localRequest, modifyRequest, pass, updateContextPath)
+import           Snap.Internal.Http.Types (Params, Request (rqContextPath, rqParams, rqPathInfo))
+import           Snap.Internal.Parsing    (urlDecode)
+#if !MIN_VERSION_base(4,8,0)
+import           Data.Monoid              (Monoid (..))
+#endif
+------------------------------------------------------------------------------
 
 ------------------------------------------------------------------------------
-import           Snap.Internal.Http.Types
-import           Snap.Internal.Parsing
-import           Snap.Internal.Types
-
-
-------------------------------------------------------------------------------
-{-|
-
-The internal data type you use to build a routing tree.  Matching is
-done unambiguously.
-
-'Capture' and 'Dir' routes can have a "fallback" route:
-
-  - For 'Capture', the fallback is routed when there is nothing to capture
-  - For 'Dir', the fallback is routed when we can't find a route in its map
-
-Fallback routes are stacked: i.e. for a route like:
-
-> Dir [("foo", Capture "bar" (Action bar) NoRoute)] baz
-
-visiting the URI foo/ will result in the "bar" capture being empty and
-triggering its fallback. It's NoRoute, so we go to the nearest parent
-fallback and try that, which is the baz action.
-
--}
+-- | The internal data type you use to build a routing tree. Matching is
+-- done unambiguously.
+--
+-- 'Capture' and 'Dir' routes can have a "fallback" route:
+--
+--   - For 'Capture', the fallback is routed when there is nothing to capture
+--   - For 'Dir', the fallback is routed when we can't find a route in its map
+--
+-- Fallback routes are stacked: i.e. for a route like:
+--
+-- > Dir [("foo", Capture "bar" (Action bar) NoRoute)] baz
+--
+-- visiting the URI foo/ will result in the "bar" capture being empty and
+-- triggering its fallback. It's NoRoute, so we go to the nearest parent
+-- fallback and try that, which is the baz action.
 data Route a m = Action ((MonadSnap m) => m a)   -- wraps a 'Snap' action
                -- captures the dir in a param
                | Capture ByteString (Route a m) (Route a m)
@@ -112,53 +115,47 @@ routeEarliestNC r n = case r of
 -- | A web handler which, given a mapping from URL entry points to web
 -- handlers, efficiently routes requests to the correct handler.
 --
+--
+-- __Usage__
+--
 -- The URL entry points are given as relative paths, for example:
 --
 -- > route [ ("foo/bar/quux", fooBarQuux) ]
 --
--- If the URI of the incoming request is
+-- If the URI of the incoming request is @\/foo\/bar\/quux@ or
+-- @\/foo\/bar\/quux\/...anything...@ then the request will be routed to
+-- @\"fooBarQuux\"@, with 'rqContextPath' set to @\"\/foo\/bar\/quux\/\"@ and
+-- 'rqPathInfo' set to @\"...anything...\"@.
 --
--- > /foo/bar/quux
---
--- or
---
--- > /foo/bar/quux/...anything...
---
--- then the request will be routed to \"@fooBarQuux@\", with 'rqContextPath'
--- set to \"@\/foo\/bar\/quux\/@\" and 'rqPathInfo' set to
--- \"@...anything...@\".
---
--- A path component within an URL entry point beginning with a colon (\"@:@\")
+-- A path component within an URL entry point beginning with a colon (@\":\"@)
 -- is treated as a /variable capture/; the corresponding path component within
 -- the request URI will be entered into the 'rqParams' parameters mapping with
 -- the given name. For instance, if the routes were:
 --
 -- > route [ ("foo/:bar/baz", fooBazHandler) ]
 --
--- Then a request for \"@\/foo\/saskatchewan\/baz@\" would be routed to
--- @fooBazHandler@ with a mapping for:
---
--- > "bar" => "saskatchewan"
---
--- in its parameters table.
+-- Then a request for @\"\/foo\/saskatchewan\/baz\"@ would be routed to
+-- @fooBazHandler@ with a mapping for @\"bar\" => \"saskatchewan\"@ in its
+-- parameters table.
 --
 -- Longer paths are matched first, and specific routes are matched before
 -- captures. That is, if given routes:
 --
 -- > [ ("a", h1), ("a/b", h2), ("a/:x", h3) ]
 --
--- a request for \"@\/a\/b@\" will go to @h2@, \"@\/a\/s@\" for any /s/ will
--- go to @h3@, and \"@\/a@\" will go to @h1@.
+-- a request for @\"\/a\/b\"@ will go to @h2@, @\"\/a\/s\"@ for any /s/ will go
+-- to @h3@, and @\"\/a\"@ will go to @h1@.
 --
--- The following example matches \"@\/article@\" to an article index,
--- \"@\/login@\" to a login, and \"@\/article\/...@\" to an article renderer.
+-- The following example matches @\"\/article\"@ to an article index,
+-- @\"\/login\"@ to a login, and @\"\/article\/...\"@ to an article renderer.
 --
--- > route [ ("article",     renderIndex)
--- >       , ("article/:id", renderArticle)
--- >       , ("login",       method POST doLogin) ]
+-- @
+-- 'route' [ (\"article\",     renderIndex)
+--       , (\"article\/:id\", renderArticle)
+--       , (\"login\",       'Snap.Core.method' POST doLogin) ]
+-- @
 --
---
--- /URL decoding/
+-- __Note: URL decoding__
 --
 -- A short note about URL decoding: path matching and variable capture are done
 -- on /decoded/ URLs, but the contents of 'rqContextPath' and 'rqPathInfo' will
@@ -168,25 +165,82 @@ routeEarliestNC r n = case r of
 -- > route [ ("a b c d/", foo ) ]
 --
 -- A request for \"@/a+b+c+d@\" will be sent to @foo@ with 'rqContextPath' set
--- to \"/a+b+c+d/\".
+-- to @\"/a+b+c+d/\"@.
 --
 -- This behaviour changed as of Snap 0.6.1; previous versions had unspecified
 -- (and buggy!) semantics here.
 --
+--
+-- __Example:__
+--
+-- @
+-- ghci> :set -XOverloadedStrings
+-- ghci> import qualified "Data.Map" as Map
+-- ghci> import qualified "Data.ByteString.Char8" as B8
+-- ghci> import "Snap.Test"
+-- ghci> :{
+-- ghci| let handler = do r \<- 'getRequest'
+-- ghci|                  'Snap.Core.writeBS' $ \"rqContextPath: \" \<> 'rqContextPath' r \<> \"\\n\"
+-- ghci|                  'Snap.Core.writeBS' $ \"rqPathInfo: \" \<> 'rqPathInfo' r \<> \"\\n\"
+-- ghci|                  'Snap.Core.writeBS' $ \"rqParams: \" \<> (B8.pack . show $ 'rqParams' r)
+-- ghci| :}
+-- ghci> 'Snap.Test.runHandler' ('Snap.Test.get' \"\/foo\/bar\" "Map.empty") ('route' [(\"foo\", handler)])
+-- HTTP\/1.1 200 OK
+-- server: Snap\/test
+-- date: Sat, 02 Aug 2014 05:16:59 GMT
+--
+-- rqContextPath: \/foo\/
+-- rqPathInfo: bar
+-- rqParams: fromList []
+-- ghci> 'Snap.Test.runHandler' ('Snap.Test.get' \"\/foo\/bar\" "Map.empty") ('route' [(\"foo\/:bar\", handler)])
+-- [...]
+--
+-- rqContextPath: \/foo\/bar\/
+-- rqPathInfo:
+-- rqParams: fromList [(\"bar\",[\"bar\"])]
+-- @
 route :: MonadSnap m => [(ByteString, m a)] -> m a
 route rts = do
   p <- getsRequest rqPathInfo
-  route' (return ()) [] (splitPath p) Map.empty rts'
+  route' (return $! ()) [] (splitPath p) Map.empty rts'
   where
     rts' = mconcat (map pRoute rts)
 {-# INLINE route #-}
 
 
 ------------------------------------------------------------------------------
--- | The 'routeLocal' function is the same as 'route'', except it doesn't
+-- | The 'routeLocal' function is the same as 'route', except it doesn't
 -- change the request's context path. This is useful if you want to route to a
 -- particular handler but you want that handler to receive the 'rqPathInfo' as
 -- it is.
+--
+-- Example:
+--
+-- @
+-- ghci> :set -XOverloadedStrings
+-- ghci> import qualified "Data.Map" as M
+-- ghci> import qualified "Data.ByteString.Char8" as B8
+-- ghci> import "Snap.Test"
+-- ghci> :{
+-- ghci| let handler = do r \<- 'getRequest'
+-- ghci|                  'Snap.Core.writeBS' $ \"rqContextPath: \" \<> 'rqContextPath' r \<> \"\\n\"
+-- ghci|                  'Snap.Core.writeBS' $ \"rqPathInfo: \" \<> 'rqPathInfo' r \<> \"\\n\"
+-- ghci|                  'Snap.Core.writeBS' $ \"rqParams: \" \<> (B8.pack . show $ 'rqParams' r)
+-- ghci| :}
+-- ghci> 'Snap.Test.runHandler' ('Snap.Test.get' \"\/foo\/bar\" M.empty) ('routeLocal' [(\"foo\", handler)])
+-- HTTP\/1.1 200 OK
+-- server: Snap\/test
+-- date: Sat, 02 Aug 2014 05:17:28 GMT
+--
+-- rqContextPath: \/
+-- rqPathInfo: foo\/bar
+-- ghci> 'Snap.Test.runHandler' ('Snap.Test.get' \"\/foo\/bar\" M.empty) ('routeLocal' [(\"foo\/:bar\", handler)])
+-- [...]
+--
+-- rqContextPath: \/
+-- rqPathInfo: foo\/bar
+-- rqParams: fromList [(\"bar\",[\"bar\"])]
+-- @
 routeLocal :: MonadSnap m => [(ByteString, m a)] -> m a
 routeLocal rts = do
     req    <- getRequest
@@ -233,7 +287,7 @@ route' pre !ctx _ !params (Action action) =
   where
     ctx' = B.intercalate (B.pack [c2w '/']) (reverse ctx)
     updateParams req = req
-      { rqParams = Map.unionWith (++) params (rqParams req) }
+      { rqParams = Map.unionWith (flip (++)) params (rqParams req) }
 
 route' pre !ctx [] !params (Capture _ _  fb) =
     route' pre ctx [] params fb
@@ -244,17 +298,17 @@ route' pre !ctx paths@(cwd:rest) !params (Capture p rt fb)
   where
     fallback = route' pre ctx paths params fb
     m = maybe pass
-              (\cwd' -> let params' = Map.insertWith (++) p [cwd'] params
+              (\cwd' -> let params' = Map.insertWith (flip (++)) p [cwd'] params
                         in route' pre (cwd:ctx) rest params' rt)
               (urlDecode cwd)
 
 route' pre !ctx [] !params (Dir _ fb) =
     route' pre ctx [] params fb
-route' pre !ctx (cwd:rest) !params (Dir rtm fb) = do
+route' pre !ctx paths@(cwd:rest) !params (Dir rtm fb) = do
     cwd' <- maybe pass return $ urlDecode cwd
     case H.lookup cwd' rtm of
       Just rt -> (route' pre (cwd:ctx) rest params rt) <|>
-                 (route' pre ctx (cwd:rest) params fb)
-      Nothing -> route' pre ctx (cwd:rest) params fb
+                 (route' pre ctx paths params fb)
+      Nothing -> route' pre ctx paths params fb
 
 route' _ _ _ _ NoRoute = pass
